@@ -3,11 +3,19 @@ use Mojo::Base 'Minion::Backend';
 
 use Carp 'croak';
 use Mojo::IOLoop;
-use Mojo::Pg 2.18;
+use Mojo::Pg 4.0;
 use Sys::Hostname 'hostname';
 
 has 'pg';
 
+sub broadcast {
+  my ($self, $command, $args, $ids) = (shift, shift, shift || [], shift || []);
+  return !!$self->pg->db->query(
+    q{update minion_workers set inbox = inbox || $1::jsonb
+      where (id = any ($2) or $2 = '{}')}, {json => [[$command, @$args]]}, $ids
+  )->rows;
+}
+
 sub dequeue {
   my ($self, $id, $wait, $options) = @_;
 
@@ -30,11 +38,12 @@ sub enqueue {
   my $db = $self->pg->db;
   return $db->query(
     "insert into minion_jobs
-       (args, attempts, delayed, parents, priority, queue, task)
-     values (?, ?, (now() + (interval '1 second' * ?)), ?, ?, ?, ?)
+       (args, attempts, delayed, notes, parents, priority, queue, task)
+     values (?, ?, (now() + (interval '1 second' * ?)), ?, ?, ?, ?, ?)
      returning id", {json => $args}, $options->{attempts} // 1,
-    $options->{delay} // 0, $options->{parents} || [],
-    $options->{priority} // 0, $options->{queue} // 'default', $task
+    $options->{delay} // 0, {json => $options->{notes} || {}},
+    $options->{parents} || [], $options->{priority} // 0,
+    $options->{queue} // 'default', $task
   )->hash->{id};
 }
 
@@ -44,11 +53,12 @@ sub finish_job { shift->_update(0, @_) }
 sub job_info {
   shift->pg->db->query(
     'select id, args, attempts,
-       array(select id from minion_jobs where j.id = any(parents)) as children,
+       array(select id from minion_jobs where parents @> ARRAY[j.id])
+         as children,
        extract(epoch from created) as created,
        extract(epoch from delayed) as delayed,
-       extract(epoch from finished) as finished, parents, priority, queue,
-       result, extract(epoch from retried) as retried, retries,
+       extract(epoch from finished) as finished, notes, parents, priority,
+       queue, result, extract(epoch from retried) as retried, retries,
        extract(epoch from started) as started, state, task, worker
      from minion_jobs as j where id = ?', shift
   )->expand->hash;
@@ -74,25 +84,52 @@ sub list_workers {
     ->arrays->map(sub { $self->worker_info($_->[0]) })->to_array;
 }
 
+sub lock {
+  my ($self, $name, $duration, $options) = (shift, shift, shift, shift // {});
+  return !!$self->pg->db->query('select * from minion_lock(?, ?, ?)',
+    $name, $duration, $options->{limit} || 1)->array->[0];
+}
+
 sub new {
   my $self = shift->SUPER::new(pg => Mojo::Pg->new(@_));
 
+  my $db = Mojo::Pg->new(@_)->db;
   croak 'PostgreSQL 9.5 or later is required'
-    if Mojo::Pg->new(@_)->db->dbh->{pg_server_version} < 90500;
-  my $pg = $self->pg->auto_migrate(1)->max_connections(1);
-  $pg->migrations->name('minion')->from_data;
+    if $db->dbh->{pg_server_version} < 90500;
+  $db->disconnect;
+
+  $self->pg->auto_migrate(1)->migrations->name('minion')->from_data;
 
   return $self;
 }
 
-sub register_worker {
-  my ($self, $id) = @_;
+sub note {
+  my ($self, $id, $key, $value) = @_;
+  return !!$self->pg->db->query(
+    'update minion_jobs set notes = jsonb_set(notes, ?, ?, true) where id = ?',
+    [$key], {json => $value}, $id
+  )->rows;
+}
 
-  return shift->pg->db->query(
-    "insert into minion_workers (id, host, pid)
-     values (coalesce(?, nextval('minion_workers_id_seq')), ?, ?)
-     on conflict(id) do update set notified = now()
-     returning id", $id, $self->{host} //= hostname, $$
+sub receive {
+  my $array = shift->pg->db->query(
+    "update minion_workers as new set inbox = '[]'
+     from (select id, inbox from minion_workers where id = ? for update) as old
+     where new.id = old.id and old.inbox != '[]'
+     returning old.inbox", shift
+  )->expand->array;
+  return $array ? $array->[0] : [];
+}
+
+sub register_worker {
+  my ($self, $id, $options) = (shift, shift, shift || {});
+
+  return $self->pg->db->query(
+    q{insert into minion_workers (id, host, pid, status)
+      values (coalesce($1, nextval('minion_workers_id_seq')), $2, $3, $4)
+      on conflict(id) do update set notified = now(), status = $4
+      returning id}, $id, $self->{host} //= hostname, $$,
+    {json => $options->{status} // {}}
   )->hash->{id};
 }
 
@@ -123,40 +160,33 @@ sub repair {
   )->hashes;
   $fail->each(sub { $self->fail_job(@$_{qw(id retries)}, 'Worker went away') });
 
-  # Jobs with missing parents (can't be retried)
-  $db->query(
-    "update minion_jobs as j
-     set finished = now(), result = to_json('Parent went away'::text),
-       state = 'failed'
-     where parents <> '{}' and cardinality(parents) <> (
-       select count(*) from minion_jobs where id = any(j.parents)
-     ) and state = 'inactive'"
-  );
-
   # Old jobs with no unresolved dependencies
   $db->query(
     "delete from minion_jobs as j
-     where finished < now() - interval '1 second' * ? and not exists (
+     where finished <= now() - interval '1 second' * ? and not exists (
        select 1 from minion_jobs
-       where j.id = any(parents) and state <> 'finished'
+       where parents @> ARRAY[j.id] and state != 'finished'
      ) and state = 'finished'", $minion->remove_after
   );
 }
 
-sub reset { shift->pg->db->query('truncate minion_jobs, minion_workers') }
+sub reset {
+  shift->pg->db->query(
+    'truncate minion_jobs, minion_locks, minion_workers restart identity');
+}
 
 sub retry_job {
   my ($self, $id, $retries, $options) = (shift, shift, shift, shift || {});
 
   return !!$self->pg->db->query(
     "update minion_jobs
-     set delayed = (now() + (interval '1 second' * ?)),
+     set attempts = coalesce(?, attempts),
+       delayed = (now() + (interval '1 second' * ?)),
        priority = coalesce(?, priority), queue = coalesce(?, queue),
        retried = now(), retries = retries + 1, state = 'inactive'
      where id = ? and retries = ?
-       and state in ('inactive', 'failed', 'finished')
-     returning 1", $options->{delay} // 0, @$options{qw(priority queue)}, $id,
-    $retries
+     returning 1", $options->{attempts}, $options->{delay} // 0,
+    @$options{qw(priority queue)}, $id, $retries
   )->rows;
 }
 
@@ -164,22 +194,32 @@ sub stats {
   my $self = shift;
 
   my $stats = $self->pg->db->query(
-    "select state::text || '_jobs', count(*) from minion_jobs group by state
-     union all
-     select 'delayed_jobs', count(*) from minion_jobs
-     where (delayed > now() or parents <> '{}') and state = 'inactive'
-     union all
-     select 'inactive_workers', count(*) from minion_workers
-     union all
-     select 'active_workers', count(distinct worker) from minion_jobs
-     where state = 'active'"
-  )->arrays->reduce(sub { $a->{$b->[0]} = $b->[1]; $a }, {});
+    "select count(*) filter (where state = 'inactive') as inactive_jobs,
+       count(*) filter (where state = 'active') as active_jobs,
+       count(*) filter (where state = 'failed') as failed_jobs,
+       count(*) filter (where state = 'finished') as finished_jobs,
+       count(*) filter (where state = 'inactive'
+         and (delayed > now() or parents != '{}')) as delayed_jobs,
+       count(distinct worker) filter (where state = 'active') as active_workers,
+       (select case when is_called then last_value else 0 end
+         from minion_jobs_id_seq) as enqueued_jobs,
+       (select count(*) from minion_workers) as inactive_workers
+     from minion_jobs"
+  )->hash;
   $stats->{inactive_workers} -= $stats->{active_workers};
-  $stats->{"${_}_jobs"} ||= 0 for qw(inactive active failed finished);
 
   return $stats;
 }
 
+sub unlock {
+  !!shift->pg->db->query(
+    'delete from minion_locks where id = (
+       select id from minion_locks
+       where expires > now() and name = ? order by expires limit 1
+     ) returning 1', shift
+  )->rows;
+}
+
 sub unregister_worker {
   shift->pg->db->query('delete from minion_workers where id = ?', shift);
 }
@@ -189,10 +229,10 @@ sub worker_info {
     "select id, extract(epoch from notified) as notified, array(
        select id from minion_jobs
        where state = 'active' and worker = minion_workers.id
-     ) as jobs, host, pid, extract(epoch from started) as started
+     ) as jobs, host, pid, status, extract(epoch from started) as started
      from minion_workers
      where id = ?", shift
-  )->hash;
+  )->expand->hash;
 }
 
 sub _try {
@@ -203,15 +243,17 @@ sub _try {
      set started = now(), state = 'active', worker = ?
      where id = (
        select id from minion_jobs as j
-       where delayed <= now() and (parents = '{}' or cardinality(parents) = (
-         select count(*) from minion_jobs
-         where id = any(j.parents) and state = 'finished'
-       )) and queue = any (?) and state = 'inactive' and task = any (?)
-       order by priority desc, created
+       where delayed <= now() and id = coalesce(?, id)
+         and (parents = '{}' or not exists (
+           select 1 from minion_jobs
+           where id = any (j.parents)
+             and state in ('inactive', 'active', 'failed')
+         )) and queue = any (?) and state = 'inactive' and task = any (?)
+       order by priority desc, id
        limit 1
        for update skip locked
      )
-     returning id, args, retries, task", $id,
+     returning id, args, retries, task", $id, $options->{id},
     $options->{queues} || ['default'], [keys %{$self->minion->tasks}]
   )->expand->hash;
 }
@@ -271,6 +313,14 @@ L<Mojo::Pg> object used to store all data.
 L<Minion::Backend::Pg> inherits all methods from L<Minion::Backend> and
 implements the following new ones.
 
+=head2 broadcast
+
+  my $bool = $backend->broadcast('some_command');
+  my $bool = $backend->broadcast('some_command', [@args]);
+  my $bool = $backend->broadcast('some_command', [@args], [$id1, $id2, $id3]);
+
+Broadcast remote control command to one or more workers.
+
 =head2 dequeue
 
   my $job_info = $backend->dequeue($worker_id, 0.5);
@@ -283,6 +333,12 @@ These options are currently available:
 
 =over 2
 
+=item id
+
+  id => '10023'
+
+Dequeue a specific job.
+
 =item queues
 
   queues => ['important']
@@ -344,7 +400,13 @@ L<Minion/"backoff"> after the first attempt, defaults to C<1>.
 
   delay => 10
 
-Delay job for this many seconds (from now).
+Delay job for this many seconds (from now), defaults to C<0>.
+
+=item notes
+
+  notes => {foo => 'bar', baz => [1, 2, 3]}
+
+Hash reference with arbitrary metadata for this job.
 
 =item parents
 
@@ -357,7 +419,7 @@ transitioned to the state C<finished> before it can be processed.
 
   priority => 5
 
-Job priority, defaults to C<0>.
+Job priority, defaults to C<0>. Jobs with a higher priority get performed first.
 
 =item queue
 
@@ -439,6 +501,12 @@ Epoch time job was delayed to.
 
 Epoch time job was finished.
 
+=item notes
+
+  notes => {foo => 'bar', baz => [1, 2, 3]}
+
+Hash reference with arbitrary metadata for this job.
+
 =item parents
 
   parents => ['10023', '10024', '10025']
@@ -538,19 +606,67 @@ List only jobs for this task.
 
 Returns the same information as L</"worker_info"> but in batches.
 
+=head2 lock
+
+  my $bool = $backend->lock('foo', 3600);
+  my $bool = $backend->lock('foo', 3600, {limit => 20});
+
+Try to acquire a named lock that will expire automatically after the given
+amount of time in seconds.
+
+These options are currently available:
+
+=over 2
+
+=item limit
+
+  limit => 20
+
+Number of shared locks with the same name that can be active at the same time,
+defaults to C<1>.
+
+=back
+
 =head2 new
 
   my $backend = Minion::Backend::Pg->new('postgresql://postgres@/test');
+  my $backend = Minion::Backend::Pg->new(Mojo::Pg->new);
 
 Construct a new L<Minion::Backend::Pg> object.
 
+=head2 note
+
+  my $bool = $backend->note($job_id, foo => 'bar');
+
+Change a metadata field for a job.
+
+=head2 receive
+
+  my $commands = $backend->receive($worker_id);
+
+Receive remote control commands for worker.
+
 =head2 register_worker
 
   my $worker_id = $backend->register_worker;
   my $worker_id = $backend->register_worker($worker_id);
+  my $worker_id = $backend->register_worker(
+    $worker_id, {status => {queues => ['default', 'important']}});
 
 Register worker or send heartbeat to show that this worker is still alive.
 
+These options are currently available:
+
+=over 2
+
+=item status
+
+  status => {queues => ['default', 'important']}
+
+Hash reference with whatever status information the worker would like to share.
+
+=back
+
 =head2 remove_job
 
   my $bool = $backend->remove_job($job_id);
@@ -574,18 +690,24 @@ Reset job queue.
   my $bool = $backend->retry_job($job_id, $retries);
   my $bool = $backend->retry_job($job_id, $retries, {delay => 10});
 
-Transition from C<failed> or C<finished> state back to C<inactive>, already
-C<inactive> jobs may also be retried to change options.
+Transition job back to C<inactive> state, already C<inactive> jobs may also be
+retried to change options.
 
 These options are currently available:
 
 =over 2
 
+=item attempts
+
+  attempts => 25
+
+Number of times performing this job will be attempted.
+
 =item delay
 
   delay => 10
 
-Delay job for this many seconds (from now).
+Delay job for this many seconds (from now), defaults to C<0>.
 
 =item priority
 
@@ -631,6 +753,13 @@ Number of jobs in C<inactive> state that are scheduled to run at specific time
 in the future or have unresolved dependencies. Note that this field is
 EXPERIMENTAL and might change without warning!
 
+=item enqueued_jobs
+
+  enqueued_jobs => 100000
+
+Rough estimate of how many jobs have ever been enqueued. Note that this field is
+EXPERIMENTAL and might change without warning!
+
 =item failed_jobs
 
   failed_jobs => 100
@@ -657,6 +786,12 @@ Number of workers that are currently not processing a job.
 
 =back
 
+=head2 unlock
+
+  my $bool = $backend->unlock('foo');
+
+Release a named lock.
+
 =head2 unregister_worker
 
   $backend->unregister_worker($worker_id);
@@ -706,6 +841,12 @@ Process id of worker.
 
 Epoch time worker was started.
 
+=item status
+
+  status => {queues => ['default', 'important']}
+
+Hash reference with whatever status information the worker would like to share.
+
 =back
 
 =head1 SEE ALSO
@@ -773,7 +914,6 @@ drop type if exists minion_state;
 
 -- 8 up
 alter table minion_jobs add constraint args check(jsonb_typeof(args) = 'array');
-create index on minion_jobs (state, priority desc, created);
 
 -- 9 up
 create or replace function minion_jobs_notify_workers() returns trigger as $$
@@ -797,4 +937,48 @@ drop trigger if exists minion_jobs_notify_workers_trigger on minion_jobs;
 drop function if exists minion_jobs_notify_workers();
 
 -- 10 up
-alter table minion_jobs add column parents bigint[] default '{}'::bigint[];
+alter table minion_jobs add column parents bigint[] not null default '{}';
+
+-- 11 up
+create index on minion_jobs (state, priority desc, id);
+
+-- 12 up
+alter table minion_workers add column inbox jsonb
+  check(jsonb_typeof(inbox) = 'array') not null default '[]';
+
+-- 15 up
+alter table minion_workers add column status jsonb
+  check(jsonb_typeof(status) = 'object') not null default '{}';
+
+-- 16 up
+create index on minion_jobs using gin (parents);
+create table if not exists minion_locks (
+  id      bigserial not null primary key,
+  name    text not null,
+  expires timestamp with time zone not null
+);
+create function minion_lock(text, int, int) returns bool as $$
+declare
+  new_expires timestamp with time zone = now() + (interval '1 second' * $2);
+begin
+  delete from minion_locks where expires < now();
+  lock table minion_locks in exclusive mode;
+  if (select count(*) >= $3 from minion_locks where name = $1) then
+    return false;
+  end if;
+  if new_expires > now() then
+    insert into minion_locks (name, expires) values ($1, new_expires);
+  end if;
+  return true;
+end;
+$$ language plpgsql;
+
+-- 16 down
+drop function if exists minion_lock(text, int, int);
+drop table if exists minion_locks;
+
+-- 17 up
+alter table minion_jobs add column notes jsonb
+  check(jsonb_typeof(notes) = 'object') not null default '{}';
+alter table minion_locks set unlogged;
+create index on minion_locks (name, expires);
